/*
 * Ind_EvolvingRobot.hpp
 *
 *  Created on: Mar 17, 2015
 *      Author: Joost Huizinga
 */

#ifndef IND_EVOLVINGROBOT_HPP_
#define IND_EVOLVINGROBOT_HPP_

//Standard includes
#include <map>

//Bullet includes
#include <btBulletDynamicsCommon.h>

//Local includes
#include "Ind_CPPN.hpp"

typedef std::vector<int> coord_t;

/**
 * A robot that can evolve its morphology based on blocks.
 */
class EvolvingRobot
{
private:


public:
    //Miscellaneous
    static const double motorMaxImpulse;
    static const double motorSpeedModifier;
    static const double cppn_scale;
    static const btScalar block_weight;
    static const btScalar friction;
    static const btScalar actuator_limits;
    static const btScalar max_frequency;

    static const int maxBlocks = 10;

    //CPPN input indices
    static const size_t x_input = 0;
    static const size_t y_input = 1;
    static const size_t z_input = 2;
    static const size_t side_input = 3;
    static const size_t x_dist_input = 4;
    static const size_t y_dist_input = 5;
    static const size_t z_dist_input = 6;

    //CPPN output indices
    static const size_t block_output = 0;
    static const size_t phaseShift_output = 1;
    static const size_t frequency_output = 2;
    static const size_t axis_output = 3;

    //Debug variable
    bool m_debug;

    //Sides
    enum side_t{
        leftSide = 0,
        rightSide,
        frontSide,
        backSide,
        topSide,
        bottomSide,
        noSide
    };



    /**
     * Constructor that creates an empty robot that will be created later.
     */
    EvolvingRobot ()
        : m_ownerWorld (0),
          m_block (0),
          m_genome(7, 4),
          m_timer(0)
    {
        m_sides.push_back(leftSide);
        m_sides.push_back(backSide);
        m_sides.push_back(bottomSide);
        m_sides.push_back(rightSide);
        m_sides.push_back(frontSide);
        m_sides.push_back(topSide);
        m_genome.setWeightMutationRate(0.1);
        m_genome.setNeuronMutationRate(0.1);
        m_debug =  false;
    }

    /**
     * Destructor, removes all parts of the robot from the world and destroys them.
     */
    virtual ~EvolvingRobot ()
    {
        if(m_ownerWorld){
            int i;

            // Remove all constraints
            for ( i = 0; i < m_joints.size(); ++i)
            {
                m_ownerWorld->removeConstraint(m_joints[i]);
                delete m_joints[i];
            }
            m_joints.clear();

            // Remove all bodies and shapes
            for ( i = 0; i < m_bodies.size(); ++i)
            {
                m_ownerWorld->removeRigidBody(m_bodies[i]);
                delete m_bodies[i]->getMotionState();
                delete m_bodies[i];
            }
            m_bodies.clear();
        }
        delete m_block;
    }


    /**
     * Function that builds the robot, given a world and a position.
     *
     * Implement such that it is build according to the CPPN genome.
     *
     * The first thing this function has to do is to store the world pointer,
     * and create the basic block shape that it will consist of, like this:
     *  m_ownerWorld = ownerWorld;
     *  m_block = new btBoxShape(btVector3(0.1, 0.1, 0.1));
     *
     *
     * Next, it is time to create a set of coordinates that should contain blocks.
     * The type of a coordinate is:
     *  typedef std::vector<int> coord_t;
     * To make sure the coordinate works, the vector should always contains 3 integers.
     *
     * To create the set of coordinates, start with an empty set of build coordinates
     * and an empty 'frontier':
     *  std::set<coord_t> build;
     *  std::list<coord_t> frontier;
     *
     * Next, add the starting coordinate (0,0,0) to the frontier.
     * Hint: coord_t(3,0) is an easy way to create a vector with 3 zeros
     *
     * Informally, you should now start a breadth first search from the
     * starting coordinate, navigating only those coordinates for which the
     * cppn genome returns a value greater than zero, and add all navigated
     * coordinates to the build set until we run out of blocks to place.
     *
     * More precisely: while the frontier is not empty, and the number of coordinates
     * in our build set is smaller than maxBlocks, do the following:
     * 1. retrieve and pop the first item from the frontier
     * 2. if it is not in the build set, add it to the build set and continue with step 3.
     *    if it is in the build set, skip it, and go back to step 1.
     * 3. for all 6 possible adjacent coordinates of the popped coordinate:
     *   - if our genome indicates that a block should be present
     *     (output > 0 given the coordinate of the block and the noSide constant):
     *     add the block the back of the frontier.
     *     To activate the CPPN, call the _activateCppn method (see below).
     *     To get the result from the cppn, use the following line of code:
     *
     *
     * You can use the following helper functions and variables:
     * - m_sides:           A vector containing the enumerator for all sides.
     *                      Useful for iterating over all sides.
     * - _getAdjacentCoord(coord_t coordinate, side_t side):
     *                      Returns an adjacent coordinate, given a coordinate
     *                      and a side (from, for example, m_sides)
     * - _activateCppn(coord_t coordinate, side_t side):
     *                      Activates the CPPN for a giving coordinate and a given side.
     *                      For this method, pass the noSide constant as a side.
     *
     *
     * Once the set of build coordinates is complete you need to build the robot.
     * First, make sure the robot is not build into the ground by updating the positionOffset.
     * Find the smallest y value in the set of build coordinates, and update the
     * y coordinate in positionOffset to accommodate for all the blocks.
     * Note that, because the coordinates are in integers, you will have to multiply
     * by the size of a block (0.2), before updating the positionOffset.
     * Also note that a block is 0.2 by 0.2 by 0.2, and placing a block at (0, 0, 0)
     * will put that block into the ground by 0.1.
     *
     * You can use the following functions:
     * - positionOffset.setY(btScalar y):
     *       Sets the y offset.
     *
     *
     * Now that we have the correct offset, add a a block at every coordinate in the build set.
     *
     * Use the following function:
     * - _addBlock(coord_t coordinate, btVector3 offset):
     *       adds a block at the indicated coordinate, given an offset.
     *
     *
     * Lastly, we have to connect all blocks at three sides: left side, bottom side, and back side.
     * The enumerates for these sides are: leftSide, bottomSide, and backSide.
     *
     * Use the following function:
     * - _connect(coord_t coordinate, side_t side):
     *       connects a coordinate at a specific side, if possible.
     *
     */
    void build(btDynamicsWorld* ownerWorld, btVector3 positionOffset){
        /* TODO: YOUR CODE HERE */

    }



    /**
     * Returns the center of mass position for the main body of the robot.
     *
     * Returns (0,0,0) if the robot has not been build.
     */
    btVector3 getPosition(){
        if(m_ownerWorld && m_bodies.size() > 0){
            return m_bodies[0]->getCenterOfMassPosition();
        } else {
            return btVector3(0,0,0);
        }
    }

    /**
     * Returns a vector with the positions of all blocks that this robot consists of.
     *
     * Returns an empty vector if the robot is not build yet.
     */
    std::vector<btVector3> getPositions(){
        std::vector<btVector3> result;
        for(size_t i=0; i<m_bodies.size(); ++i){
            result.push_back(m_bodies[i]->getCenterOfMassPosition());
        }
        return result;
    }


    /**
     * Debug print function.
     *
     * Does not print any limits if the robot has not been build.
     */
    void printLimits(){
        std::cout << "Printing limits" << std::endl;
        for(size_t jointIndex=0; jointIndex<m_joints.size(); ++jointIndex){
            std::cout << m_joints[jointIndex]->getLowerLimit() << " " << m_joints[jointIndex]->getUpperLimit() << std::endl;
        }
    }

    /**
     * Performs one step of the robot.
     */
    void step(double timeStep){
        m_timer += timeStep;
        double motorCommand;
        for(int i = 0; i < m_joints.size(); ++i){
            motorCommand = (sin(m_timer*m_motorFrequency[i] + m_motorPhaseShift[i])+1)/2;
            m_motorCommands[i] = motorCommand;
        }
    }

    /**
     * Executes the motor commands stored in m_motorCommands.
     */
    void executeMotorCommand(){
        for(int i = 0; i < m_motorCommands.size(); ++i){
            double motorCommand = m_motorCommands[i];
            _actuateJoint(i, motorCommand);
        }
    }

    /**
     * Mutates this individual.
     *
     * Simply mutates the underlying cppn.
     */
    void mutate(){
        m_genome.mutate();
    }

    /**
     * Randomizes this individual.
     *
     * Simply randomized the underlying cppn.
     */
    void randomize(){
        m_genome.randomize();
    }

private:
    //A pointer to the world, once we are build
    btDynamicsWorld* m_ownerWorld;

    //The shape of the basic block that this robot will consist of
    btCollisionShape* m_block;

    //A vector with bodies
    std::vector<btRigidBody*> m_bodies;

    //A vector of joints
    std::vector<btHingeConstraint*> m_joints;

    //The vector of motor commands, executed on a call to executeMotorCommand
    std::vector<double> m_motorCommands;

    //The frequency for each joint (evolved)
    std::vector<double> m_motorFrequency;

    //The phase shift for each joint (evolved)
    std::vector<double> m_motorPhaseShift;

    //Convenience construct: the component map maps coordinates to components
    std::map<coord_t, btRigidBody*> m_componentMap;

    //Convenience construct: the sides vector makes it easy to iterate over all sides
    std::vector<side_t> m_sides;

    //The genome used to build this individual
    CPPN m_genome;

    //An internal timer to feed to the sine-wave operating the robot
    btScalar m_timer;


    /**
     * Connects a block at the indicated coordinate to a block at the indicated side, if present.
     *
     * Implement such that the block at thisCoord gets connected to the block at side, if there
     * is indeed a block present at that side. Also set the phaseShift, frequency and axis for this
     * joint, according to the cppn genome.
     *
     *
     * First, determine whether there is an adjacent block at the indicated side. If there is not
     * a block at the indicated side, return immediately.
     *
     * To get the coordinate adjacent to thisCoord at side, call:
     * - _getAdjacentCoord(coord_t coordinate, side_t side):
     *       Returns an adjacent coordinate, given a coordinate and a side.
     *
     * To check whether a block is present at a certain coordinate, call:
     * - m_componentMap.count(coordinate) > 0
     *
     *
     * Next, set the phase shift, frequency and axis of joint we are about the create.
     *
     * First activate the cppn through:
     * - _activateCppn(coord_t coordinate, side_t side):
     *     Activates the CPPN for a giving coordinate and a given side.
     * Remember to use thisCoordinate, and not the adjacent coordinate, as the first argument.
     *
     * To set the phase shift, first query the CPPN by calling:
     * - m_genome.getOutputValue(phaseShift_output)
     * Then, because the output of the cppn will be in [-1, 1], multiply the value
     * by 3.14159 to get a decent range.
     * Lastly, push_back the result to m_motorPhaseShift.
     *
     * To set the frequency, call:
     * - m_genome.getOutputValue(frequency_output)
     * Scale the frequency to useful values by first adding 1 (so the result is in [0, 2]),
     * and then multiplying the result by 16 (so the result is in [0, 32]).
     * Push_back the result to m_motorFrequency.
     *
     * To set the axis, first define a btVector3 called axis:
     * - btVector3 axis
     * Then call:
     * - m_genome.getOutputValue(axis_output)
     * to query the CPPN. Now, set the axis as follows:
     * If the queried value is > 0.33, let the axis be: btVector3(1, 0, 0)
     * If the queried value is <= 0.33 but > -0.33, let the axis be: btVector3(0, 1, 0)
     * If the queried value is <= -0.33, let the axis be: btVector3(0, 0, 1)
     *
     *
     * Lastly, build the actual joint by calling:
     * - _addJoint(btRigidBody* body1, btRigidBody* body2, btVector3 axis, side_t side)
     * To retrieve the rigid bodies of the two blocks, you can use the the coordinate map
     * as follows:
     * - btRigidBody* block = m_componentMap[coordinate];
     * where the coordinate of the first body should be thisCoord, and the coordinate of
     * the second body should be the relevant adjacent coordinate.
     * The axis should be the axis you determined in the step above, and the side should
     * be the side passed to this function.
     *
     * Now this function should connect two adjacent blocks if both are present.
     */
    void _connect(const coord_t& thisCoord, side_t side){
        /* TODO: YOUR CODE HERE */
 
    }

    /**
     * Activates the cppn for the given coordinate and the given side.
     *
     * Implement such that it activates the cppn (m_genome) for the given coordinate and the given side.
     *
     * Use:
     * - m_genome.setValue(size_t index, double value)
     * to set a particular input of the cppn.
     *
     * The following constants are available to indicate inputs:
     * - x_input
     * - y_input
     * - z_input
     * - side_input
     * - x_dist_input
     * - y_dist_input
     * - z_dist_input
     *
     * The first three values of the cppn should simply be the first three values in coord, scaled by
     * the cppn_scale variable (e.g. for the first input: m_genome.setValue(x_input, coord[0]*cppn_scale)).
     *
     * For the fourth input, named the side input, provide the following values:
     * - if the side is leftSide or rightSide: -1
     * - if the side is topSide, bottomSide, or noSide: 0
     * - if the side is frontSide or backSide: 1
     *
     * For the last three inputs (x_dist_input, y_dist_input, z_dist_input), provide the absolute difference
     * between the coordinate and 0, scaled by the cppn_scale
     * (e.g. m_genome.setValue(x_dist_input, abs(0-coord[0])*cppn_scale)).
     *
     * Activate the network by calling:
     * - m_genome.step();
     * a number of times, equal to the depth of the network.
     *
     * Because incoming values are reset each step, we will have to call the setValue method before
     * EVERY step, like this:
     *   for(int i=0; i < ...; ++i){
     *       m_genome.setValue(x_input, coord[0]*cppn_scale));
     *       m_genome.setValue(..., ...));
     *       ...
     *       m_genome.setValue(..., ...));
     *       m_genome.step();
     *   }
     */
    void _activateCppn(const coord_t& coord, side_t side){
        /* TODO: YOUR CODE HERE */
    }

    /**
     * Returns the coordinate of the block that would be adjacent to the
     * supplied coordinate in the indicated direction.
     */
    coord_t _getAdjacentCoord(const coord_t& current, side_t side){
        coord_t copy = current;
        switch(side){
        case leftSide:
            copy[0] -= 1;
            break;
        case rightSide:
            copy[0] += 1;
            break;
        case frontSide:
            copy[2] += 1;
            break;
        case backSide:
            copy[2] -= 1;
            break;
        case topSide:
            copy[1] += 1;
            break;
        case bottomSide:
            copy[1] -= 1;
            break;
        default:
            std::cerr << "Unknown side: " << side << std::endl;
        }
        return copy;
    }

    /**
     * Returns the pivot associated with the supplied side.
     */
    btVector3 _getPivot(side_t side){
        switch(side){
        case leftSide:
            return btVector3(-0.1,0,0);
        case rightSide:
            return btVector3(0.1,0,0);
        case frontSide:
            return btVector3(0,0,0.1);
        case backSide:
            return btVector3(0,0,-0.1);
        case topSide:
            return btVector3(0,0.1,0);
        case bottomSide:
            return btVector3(0,-0.1,0);
        default:
            std::cerr << "Unknown side: " << side << std::endl;
            return btVector3(0,0,0);
        }
    }

    /**
     * Adds a joint between the indicated body parts along the supplied side, with the supplied axis.
     *
     * Note: it is the callers responsibility to ensure that the body parts are actually adjacent,
     * and touching at the indicated side. Violating this pre-condition can cause the robot to explode
     * and the simulator to crash with an AABB error.
     */
    void _addJoint(btRigidBody* body1, btRigidBody* body2, const btVector3& axis, side_t side){
        btVector3 pivotInA = _getPivot(side);
        btVector3 pivotInB = -1 * pivotInA;
        if(axis.fuzzyZero()){
            btVector3 temp(0, 1, 0);
            m_joints.push_back(new btHingeConstraint(*body1, *body2, pivotInA, pivotInB, temp, temp));
            m_joints.back()->setLimit(0, 0);
        } else {
            m_joints.push_back(new btHingeConstraint(*body1, *body2, pivotInA, pivotInB, axis, axis));
            m_joints.back()->setLimit(-actuator_limits*3.14159 / 180., actuator_limits*3.14159 / 180.);
        }

        m_motorCommands.push_back(0);
        m_ownerWorld->addConstraint(m_joints.back(), true);
    }

    /**
     * Adds a block at the indicated coordinate, with the indicated offset.
     */
    void _addBlock(const coord_t& coord, btVector3 offset){
        btTransform offsetTransform;
        offsetTransform.setIdentity();
        offsetTransform.setOrigin(offset);

        btTransform transform;
        transform.setIdentity();
        transform.setOrigin(btVector3(coord[0]*0.2, coord[1]*0.2, coord[2]*0.2));
        m_bodies.push_back(_localCreateRigidBody(btScalar(1.), offsetTransform*transform, m_block));
        m_componentMap[coord] = m_bodies.back();
    }

    /**
     * Actuates the joint at index jointIndex towards the desired angle.
     *
     * Note that the joint will only move towards the desired angle,
     * there is no guarantee that the actuator will reach that angle in a single time-step.
     */
    void _actuateJoint(int jointIndex, double desiredAngle)
    {
        double range = m_joints[jointIndex]->getUpperLimit() - m_joints[jointIndex]->getLowerLimit();
        if(range < 0.0001) return;
        double desired = m_joints[jointIndex]->getLowerLimit() + range*desiredAngle;
        double diff = desired - m_joints[jointIndex]->getHingeAngle();
        diff = std::max(std::min(diff, 1.0), -1.0);
        m_joints[jointIndex]->enableAngularMotor(true, motorSpeedModifier*diff, motorMaxImpulse);
    }

    /**
     * Creates a rigid body with the supplied mass, shape and transform.
     *
     * Helper for function for creating the body parts of the robot.
     */
    btRigidBody* _localCreateRigidBody (btScalar mass, const btTransform& startTransform, btCollisionShape* shape)
    {
        bool isDynamic = (mass != 0.f);

        btVector3 localInertia(0,0,0);
        if (isDynamic)
            shape->calculateLocalInertia(mass,localInertia);

        btDefaultMotionState* myMotionState = new btDefaultMotionState(startTransform);

        btRigidBody::btRigidBodyConstructionInfo rbInfo(mass,myMotionState,shape,localInertia);
        rbInfo.m_friction = friction;
        btRigidBody* body = new btRigidBody(rbInfo);

        // Old (2016) setup
        m_ownerWorld->addRigidBody(body);
        // New (2017) setup
//        m_ownerWorld->addRigidBody(body, 2, 1);

        return body;
    }
};

//Declare these values outside of the class,
//just in case a compiler does not support in class declarations.
//This WILL give errors when there is more than one cpp file (more than one symbol found)
//But that should not happen during these assignments
const double EvolvingRobot::cppn_scale = 0.4;
const double EvolvingRobot::motorMaxImpulse = 2;
const double EvolvingRobot::motorSpeedModifier = 10;
const btScalar EvolvingRobot::block_weight = 1.0;
const btScalar EvolvingRobot::friction = 1;
const btScalar EvolvingRobot::actuator_limits = 45;
const btScalar EvolvingRobot::max_frequency = 32;



#endif /* IND_EVOLVINGROBOT_HPP_ */
